#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import print_function

from datetime import datetime as _datetime
from functools import wraps as _wraps
from plano import *

import collections as _collections
import distutils.sysconfig as _sysconfig
import inspect as _inspect
import json as _json
import multiprocessing as _multiprocessing
import os as _os
import subprocess as _subprocess
import sys as _sys
import traceback as _traceback
import types as _types

class _Object(object):
    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.name)

class Rivet(_Object):
    def __init__(self, home):
        super(Rivet, self).__init__()

        self.home = home

        self.commands = list()
        self.commands_by_name = dict()

        self.projects = list()
        self.projects_by_name = dict()

        Command(self, "info")
        Command(self, "export")
        Command(self, "build")
        Command(self, "install")
        Command(self, "uninstall")
        ReleaseCommand(self, "release")
        Command(self, "test-build")
        Command(self, "test-install")
        Command(self, "test-release")
        Command(self, "test")
        CleanCommand(self, "clean")
        EnvironmentCommand(self, "env")

    def load(self):
        for project in self.projects:
            project.load()

    def init(self):
        for project in self.projects:
            project.init()

    def get_source_dir(self, request):
        return join(request.output_dir, "source")

    def get_release_dir(self, request):
        return join(request.output_dir, "release")

    def get_build_dir(self, request):
        return join(request.output_dir, "build")

    def get_install_dir(self, request):
        return request.install_dir

    def get_install_env(self, request):
        install_dir = self.get_install_dir(request)
        bin_dir = self.get_bin_dir(request)
        sbin_dir = self.get_sbin_dir(request)
        lib_dir = self.get_lib_dir(request)
        python_dir = self.get_python_lib_dir(request)
        plat_python_dir = self.get_python_plat_lib_dir(request)

        env = dict()
        env["PATH"] = ":".join((bin_dir, sbin_dir, _os.environ["PATH"]))
        env["LD_LIBRARY_PATH"] = lib_dir
        env["PYTHONPATH"] = ":".join((plat_python_dir, python_dir))

        return env

    def get_bin_dir(self, request):
        return join(request.install_dir, "bin")

    def get_sbin_dir(self, request):
        return join(request.install_dir, "sbin")

    def get_include_dir(self, request):
        return join(request.install_dir, "include")

    def get_lib_dir(self, request):
        if _sys.maxsize > 2**32:
            return join(request.install_dir, "lib64")

        return join(request.install_dir, "lib64")

    def get_python_lib_dir(self, request):
        return _sysconfig.get_python_lib(False, False, request.install_dir)

    def get_python_plat_lib_dir(self, request):
        return _sysconfig.get_python_lib(True, False, request.install_dir)

class Request(_Object):
    def __init__(self, output_dir):
        super(Request, self).__init__()

        self.output_dir = absolute_path(output_dir)

        self.modules = list()
        self.source_revisions_by_module = dict()

        self.concurrent_jobs = _multiprocessing.cpu_count()
        self.install_dir = join(self.output_dir, "install")
        self.release_tag = None

        self.visited_tasks = set()
        self.failed_tasks = dict()

        self.date = _datetime.now().strftime("%Y%m%d")

class Command(_Object):
    def __init__(self, app, name):
        super(Command, self).__init__(name)

        self.app = app
        self.method_name = self.name.replace("-", "_")

        assert self.name not in self.app.commands_by_name

        self.app.commands.append(self)
        self.app.commands_by_name[self.name] = self

    def __call__(self, request):
        for module in request.modules:
            task = getattr(module, self.method_name)
            task(request)

class ReleaseCommand(Command):
    def __call__(self, request):
        super(ReleaseCommand, self).__call__(request)

        release_dir = self.app.get_release_dir(request)

        print("Current release files:")

        for file in find(release_dir):
            print("  {}".format(file))

class CleanCommand(Command):
    def __call__(self, request):
        super(CleanCommand, self).__call__(request)

        install_dir = self.app.get_install_dir(request)

        if install_dir.startswith(request.output_dir):
            remove(install_dir)

class EnvironmentCommand(Command):
    def __call__(self, request):
        env = self.app.get_install_env(request)

        # XXX Factor in project-specific install env settings

        for name in sorted(env):
            value = env[name]
            print("export {}={}".format(name, value))

class Task(object):
    def __init__(self, module, name):
        self.module = module
        self.name = name

        method_name = "do_{}".format(self.name.replace("-", "_"))
        self.method = getattr(self.module, method_name)

    def __repr__(self):
        return "{}({})".format(self.name, self.module.name)

    def __call__(self, request):
        if self in request.visited_tasks:
            return

        request.visited_tasks.add(self)

        notice("Starting {}", self)

        try:
            self.do_call(request)
        except KeyboardInterrupt:
            raise
        except _subprocess.CalledProcessError as e:
            request.failed_tasks[self] = e
            notice("Call failed: {}", str(e))
        except Exception as e:
            request.failed_tasks[self] = e
            error(e)

        notice("Finished {}", self)

    def do_call(self, request):
        self.method(request)

class Project(_Object):
    def __init__(self, app, name):
        super(Project, self).__init__(name)

        self.app = app

        self.modules = list()
        self.modules_by_name = dict()

        assert self.name not in self.app.projects_by_name

        self.app.projects.append(self)
        self.app.projects_by_name[self.name] = self

    def load(self):
        for module in self.modules:
            module.load()

    def init(self):
        for module in self.modules:
            module.init()

class Module(_Object):
    def __init__(self, project, name):
        super(Module, self).__init__(name)

        self.project = project
        self.app = project.app

        self.source_url = None
        self.source_branch = None
        self.source_revision = None

        self.release_tag = None

        assert name not in self.project.modules_by_name

        self.project.modules.append(self)
        self.project.modules_by_name[self.name] = self

        self.info = Task(self, "info")
        self.clean = Task(self, "clean")
        self.export = Task(self, "export")
        self.build = BuildTask(self, "build")
        self.install = InstallTask(self, "install")
        self.release = ReleaseTask(self, "release")
        self.test = TestTask(self, "test")

    def load(self):
        if self.release_tag is None:
            date = _datetime.now().strftime("%Y%m%d")
            self.release_tag = "{}-{}".format(self.source_branch, date)

    def init(self):
        pass

    def get_source_url(self, request):
        return self.source_url

    def get_source_branch(self, request):
        return self.source_branch

    def get_source_revision(self, request):
        revision = request.source_revisions_by_module.get(self)

        if revision is not None:
            return revision

        return self.source_revision

    def get_shortened_source_revision(self, request):
        return self.get_source_revision(request)

    def get_source_dir(self, request):
        return join(self.app.get_source_dir(request), self.name)

    def get_build_dir(self, request):
        return join(self.app.get_build_dir(request), self.name)

    def get_install_dir(self, request):
        return self.app.get_install_dir(request)

    def get_release_dir(self, request):
        return join(self.app.get_release_dir(request), self.name)

    def get_release_tag(self, request):
        if request.release_tag is not None:
            return request.release_tag

        source_branch = self.get_source_branch(request)
        source_revision = self.get_shortened_source_revision(request)

        return "{}-{}-{}".format(source_branch, request.date, source_revision)

    def get_revision_file(self, request):
        source_dir = self.get_source_dir(request)
        return "{}.revision".format(source_dir)

    def get_archive_stem(self, request):
        release_tag = self.get_release_tag(request)
        return "{}-{}".format(self.name, release_tag)

    def call(self, request, command, *args, **kwargs):
        kwargs["env"] = self.app.get_install_env(request)
        kwargs["shell"] = True

        call(command, *args, **kwargs)

    def call_for_output(self, request, command, *args, **kwargs):
        kwargs["env"] = self.app.get_install_env(request)
        kwargs["shell"] = True

        return call_for_output(command, *args, **kwargs)

    def do_info(self, request):
        source_url = self.get_source_url(request)
        source_branch = self.get_source_branch(request)
        source_revision = self.get_source_revision(request)
        source_dir = self.get_source_dir(request)
        build_dir = self.get_build_dir(request)
        release_tag = self.get_release_tag(request)

        print("Module '{}':".format(self.name))
        print("  source_url:        {}".format(source_url))
        print("  source_branch:     {}".format(source_branch))
        print("  source_revision:   {}".format(source_revision))
        print("  source_dir:        {}".format(source_dir))
        print("  build_dir:         {}".format(build_dir))
        print("  release_tag:       {}".format(release_tag))

    def do_clean(self, request):
        source_dir = self.get_source_dir(request)
        build_dir = self.get_build_dir(request)
        release_dir = self.get_release_dir(request)
        revision_file = self.get_revision_file(request)

        remove(source_dir)
        remove(build_dir)
        remove(release_dir)
        remove(revision_file)

    def do_release(self, request):
        source_dir = self.get_source_dir(request)
        release_dir = self.get_release_dir(request)
        archive_stem = self.get_archive_stem(request)

        archive_file = make_archive(source_dir, release_dir, archive_stem)

        self.copy_revision_file(request)
        self.generate_checksums(archive_file)

    def copy_revision_file(self, request):
        revision_file = self.get_revision_file(request)
        release_dir = self.get_release_dir(request)
        archive_stem = self.get_archive_stem(request)
        release_revision_file = "{}.revision".format \
            (join(release_dir, archive_stem))

        copy(revision_file, release_revision_file)

    def generate_checksums(self, archive_file):
        release_dir, archive_name = split(archive_file)

        with working_dir(release_dir):
            call("md5sum {} > {}.md5", archive_name, archive_name, shell=True)
            call("sha1sum {} > {}.sha1", archive_name, archive_name, shell=True)

    def do_export(self, request):
        raise NotImplementedError()

    def do_build(self, request):
        raise NotImplementedError()

    def do_install(self, request):
        pass

    def do_test(self, request):
        raise NotImplementedError()

class BuildTask(Task):
    def do_call(self, request):
        self.module.export(request)

        super(BuildTask, self).do_call(request)

class InstallTask(Task):
    def do_call(self, request):
        self.module.build(request)

        super(InstallTask, self).do_call(request)

class ReleaseTask(Task):
    def do_call(self, request):
        self.module.export(request)

        super(ReleaseTask, self).do_call(request)

class TestTask(Task):
    def do_call(self, request):
        self.module.install(request)
        self.module.release(request)

        super(TestTask, self).do_call(request)
